import { KNEX_NO_ARGS, COLUMN_TYPES, IGNORABLE_PROPS } from './constants';
import { isWhereMultiple, isWhereTuple, findKey, runQuery } from './helpers';
import { eachObj, invariant, isFunction, isObject, isString, mapObj, toArray } from './util';
import * as types from './types';
export function toKnexSchema(model, options) {
    return (table) => {
        // every property of `model.schema` is a column
        eachObj(model.schema, (descriptor, name) => {
            if (options.timestamps && ['created_at', 'updated_at'].includes(name))
                return;
            // each column's value is either its type or a descriptor
            const type = getDataType(descriptor);
            const partial = table[toKnexMethod(type)](name);
            if (isFunction(descriptor) || !isObject(descriptor))
                return;
            const props = types.ColumnDescriptor.check(descriptor);
            if ('nullable' in props) {
                if ('notNullable' in props) {
                    invariant(false, `can't set both 'nullable' & 'notNullable' - they work inversely`);
                }
                props.notNullable = !props.nullable;
            }
            eachObj(props, (value, property) => {
                if (IGNORABLE_PROPS.has(property))
                    return;
                if (KNEX_NO_ARGS.has(property)) {
                    props[property] && partial[property]();
                }
                else {
                    props[property] && partial[property](value);
                }
            });
        });
        for (const key of Object.keys(options)) {
            const value = options[key];
            if (key === 'timestamps' && options.timestamps) {
                table.timestamps(false, true);
                // tslint:disable-next-line:no-floating-promises
                createTimestampTrigger(model);
            }
            else if (key === 'index') {
                createIndices(table, value);
            }
            else {
                table[key](value);
            }
        }
    };
}
function createIndices(table, value) {
    if (isString(value)) {
        table.index([value]);
    }
    else if (Array.isArray(value)) {
        if (value.every(isString)) {
            table.index(value);
        }
        value.forEach(columns => table.index(columns));
    }
    else if (isObject(value)) {
        for (const [indexName, columns] of Object.entries(value)) {
            table.index(toArray(columns), indexName);
        }
    }
}
export var TriggerEvent;
(function (TriggerEvent) {
    TriggerEvent["Insert"] = "insert";
    TriggerEvent["Update"] = "update";
    TriggerEvent["Delete"] = "delete";
})(TriggerEvent || (TriggerEvent = {}));
export async function createTrigger(model, event) {
    const keys = Object.keys(model.schema);
    const tableName = `${model.name}_returning_temp`;
    const triggerName = `on_${event}_${model.name}`;
    const fieldPrefix = event === TriggerEvent.Delete ? 'old.' : 'new.';
    const fieldReferences = keys.map(k => fieldPrefix + k).join(', ');
    await Promise.all([
        runQuery(model.ctx, model.ctx.knex.raw(`
      create table if not exists ${tableName} (
        ${keys.join(', ')}
      )
    `), { model, internal: true }),
        runQuery(model.ctx, model.ctx.knex.raw(`
      create trigger if not exists ${triggerName}
        after ${event} on ${model.name}
        begin
          insert into ${tableName} select ${fieldReferences};
        end
    `), { model, internal: true })
    ]);
    let query = model.ctx.knex(tableName);
    if (event === TriggerEvent.Insert) {
        // tslint:disable-next-line:semicolon
        ;
        query = query.first();
    }
    const cleanup = () => {
        return Promise.all([
            model.ctx.knex.raw(`drop table if exists ${tableName}`),
            model.ctx.knex.raw(`drop trigger if exists ${triggerName}`)
        ].map(query => runQuery(model.ctx, query, { model, internal: true })));
    };
    return [query, cleanup];
}
export function createTimestampTrigger(model, column = 'updated_at') {
    const { key, hasIncrements } = findKey(model.schema);
    if (!key && !hasIncrements) {
        // there's no way to uniquely identify the updated record
        return Promise.resolve();
    }
    const query = model.ctx.knex.raw(`
    create trigger if not exists on_update_${model.name}_timestamp
      after update on ${model.name}
      begin
        update ${model.name}
          set ${column} = current_timestamp
          where ${key} = old.${key};
      end
  `);
    return runQuery(model.ctx, query, { model, internal: true });
}
export function castValue(value) {
    const type = typeof value;
    if (type === 'number' || type === 'string') {
        return value;
    }
    if (type === 'boolean')
        return Number(value);
    if (Array.isArray(value) || isObject(value)) {
        return JSON.stringify(value);
    }
    return value;
}
export function normalizeSchema(schema, options) {
    const keys = Object.keys(schema);
    invariant(keys.length > 0, 'model schemas cannot be empty');
    const result = {};
    for (const key of keys) {
        const descriptor = schema[key];
        const type = typeof descriptor;
        result[key] = type === 'function' || type === 'string'
            ? { type: descriptor }
            : descriptor;
    }
    if (options.timestamps) {
        // tslint:disable-next-line:semicolon
        ;
        result.created_at = { type: Date };
        result.updated_at = { type: Date };
    }
    return result;
}
function getDataType(property) {
    let type = property;
    if (isFunction(property)) {
        type = property.name;
    }
    else if (isObject(property)) {
        type = isFunction(property.type)
            ? property.type.name
            : property.type;
    }
    if (isString(type)) {
        const lower = type.toLowerCase();
        if (!COLUMN_TYPES.has(lower)) {
            return 'string';
        }
        return lower;
    }
    return invariant(false, `column type must be of type string`);
}
export function toKnexMethod(type) {
    switch (type) {
        case 'string':
        case 'array':
        case 'object':
        case 'json':
            return 'text';
        case 'number':
        case 'boolean':
            return 'integer';
        case 'date':
            return 'dateTime';
        case 'increments':
            return 'increments';
        default:
            return invariant(false, `invalid column type definition: ${type}`);
    }
}
export function toInputType(type, value) {
    switch (type) {
        case 'string':
            return String(value);
        case 'array':
        case 'object':
        case 'json':
            return JSON.stringify(value);
        case 'number':
        case 'boolean':
        case 'increments':
            return Number(value);
        case 'date':
            return value.toISOString();
        default:
            return invariant(false, `invalid type on insert to database: ${type}`);
    }
}
export function toReturnType(type, value) {
    switch (type) {
        case 'string':
            return String(value);
        case 'array':
        case 'object':
        case 'json':
            return JSON.parse(value);
        case 'number':
        case 'increments':
            return Number(value);
        case 'boolean':
            return Boolean(value);
        case 'date':
            return new Date(value);
        default:
            return invariant(false, `invalid type returned from database: ${type}`);
    }
}
export class Cast {
    constructor(model) {
        this.model = model;
    }
    toDefinition(object, options) {
        if (isWhereTuple(object)) {
            const clone = object.slice();
            const valueIndex = clone.length - 1;
            clone[valueIndex] =
                this.toColumnDefinition(clone[0], clone[valueIndex], options);
            return clone;
        }
        if (isWhereMultiple(object)) {
            return object.map(clause => this.toDefinition(clause, options));
        }
        if (isObject(object)) {
            return mapObj(object, (value, column) => {
                return this.toColumnDefinition(column, value, options);
            });
        }
        return invariant(false, `invalid input type: '${typeof object}'`);
    }
    fromDefinition(object, options) {
        return mapObj(object, (value, column) => {
            return this.fromColumnDefinition(column, value, options);
        });
    }
    toColumnDefinition(column, value, options = { raw: false }) {
        const definition = this.model.schema[column];
        invariant(!(definition.notNullable && value == null), `${this.model.name}.${column} is not nullable but received nil`);
        const type = getDataType(definition);
        const cast = value !== null ? toInputType(type, value) : value;
        if (!options.raw && isFunction(definition.set)) {
            return castValue(definition.set(cast));
        }
        return cast;
    }
    fromColumnDefinition(column, value, options = { raw: false }) {
        const definition = this.model.schema[column];
        const type = getDataType(definition);
        const cast = value !== null ? toReturnType(type, value) : value;
        if (!options.raw && isFunction(definition.get)) {
            return definition.get(cast);
        }
        return cast;
    }
}
